"""
rflint - a lint-like tool for robot framework plain text files

Copyright 2014-2015 Bryan Oakley

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
from __future__ import print_function

import os
import sys
import glob
import argparse
import imp

from .common import SuiteRule, ResourceRule, TestRule, KeywordRule, GeneralRule, Rule
from .common import ERROR, WARNING, IGNORE
from .version import __version__
from .parser import RobotFactory, SuiteFile, ResourceFile
from .exceptions import UnknownRuleException

from robot.utils.argumentparser import ArgFileParser

# Used to track which files have already been imported
IMPORTED_RULE_FILES = []


class RfLint(object):
    """Robot Framework Linter"""

    def __init__(self):
        here = os.path.abspath(os.path.dirname(__file__))
        builtin_rules = os.path.join(here, "rules")
        site_rules = os.path.join(here, "site-rules")

        # mapping of class names to instances, to enable us to
        # instantiate each rule exactly once
        self._rules = {}

        for path in (builtin_rules, site_rules):
            for filename in glob.glob(path+"/*.py"):
                if filename.endswith(".__init__.py"):
                    continue
                self._load_rule_file(filename)

    @property
    def suite_rules(self):
        return self._get_rules(SuiteRule)

    @property
    def resource_rules(self):
        return self._get_rules(ResourceRule)

    @property
    def testcase_rules(self):
        return self._get_rules(TestRule)

    @property
    def keyword_rules(self):
        return self._get_rules(KeywordRule)

    @property
    def general_rules(self):
        return self._get_rules(GeneralRule)

    @property
    def all_rules(self):
        all = self.suite_rules + self.resource_rules + self.testcase_rules + self.keyword_rules + self.general_rules
        return all

    def run(self, args):
        """Parse command line arguments, and run rflint"""

        self.args = self.parse_and_process_args(args)

        if self.args.version:
            print(__version__)
            return 0

        if self.args.rulefile:
            for filename in self.args.rulefile:
                self._load_rule_file(filename)

        if self.args.list:
            self.list_rules()
            return 0

        if self.args.describe:
            self._describe_rules(self.args.args)
            return 0

        self.counts = { ERROR: 0, WARNING: 0, "other": 0}

        for filename in self.args.args:
            if not (os.path.exists(filename)):
                sys.stderr.write("rflint: %s: No such file or directory\n" % filename)
                continue
            if os.path.isdir(filename):
                self._process_folder(filename)
            else:
                self._process_file(filename)

        if self.counts[ERROR] > 0:
            return self.counts[ERROR] if self.counts[ERROR] < 254 else 255

        return 0

    def _is_valid_rule(self, rule_name):
        for rule in self.all_rules:
            if rule_name.lower() == rule.name.lower():
                return True
        return False

    def _describe_rules(self, rule_names):
        for rulename in rule_names:
            if not self._is_valid_rule(rulename):
                raise UnknownRuleException(rulename)

        requested_rules = [rule.strip().lower() for rule in rule_names]
        for rule in sorted(self.all_rules, key=lambda rule: rule.name):
            if rule.name.lower() in requested_rules or len(requested_rules) == 0:
                print(rule.name)
                for line in rule.doc.split("\n"):
                    print("    " + line)

    def _process_folder(self, path):
        if self.args.recursive:
            for root, subdirs, filenames in os.walk(path):
                self._process_files(root, filenames)
        else:
            self._process_files(path, os.listdir(path))

    def _process_files(self, folder, filenames):
        for filename in filenames:
            name, ext = os.path.splitext(filename)
            if ext.lower() in (".robot", ".txt", ".tsv", ".resource"):
                self._process_file(os.path.join(folder, filename))

    def _process_file(self, filename):
        # this is used by the reporting mechanism to know if it
        # should print the filename. Once it has been printed it
        # will be reset so that it won't get printed again until
        # we process the next file.
        self._print_filename = filename if self.args.print_filenames else None

        robot_file = RobotFactory(filename)
        for rule in self.general_rules:
            if rule.severity != IGNORE:
                rule.apply(robot_file)

        if isinstance(robot_file, SuiteFile):
            for rule in self.suite_rules:
                if rule.severity != IGNORE:
                    rule.apply(robot_file)
            for testcase in robot_file.testcases:
                for rule in self.testcase_rules:
                    if rule.severity != IGNORE:
                        rule.apply(testcase)

        if isinstance(robot_file, ResourceFile):
            for rule in self.resource_rules:
                if rule.severity != IGNORE:
                    rule.apply(robot_file)

        for keyword in robot_file.keywords:
            for rule in self.keyword_rules:
                if rule.severity != IGNORE:
                    rule.apply(keyword)

    def list_rules(self):
        """Print a list of all rules"""
        for rule in sorted(self.all_rules, key=lambda rule: rule.name):
            print(rule)
            if self.args.verbose:
                for line in rule.doc.split("\n"):
                    print("    ", line)

    def report(self, linenumber, filename, severity, message, rulename, char):
        """Report a rule violation"""

        if self._print_filename is not None:
            # we print the filename only once. self._print_filename
            # will get reset each time a new file is processed.
            print("+ " + self._print_filename)
            self._print_filename = None

        if severity in (WARNING, ERROR):
            self.counts[severity] += 1
        else:
            self.counts["other"] += 1

        if sys.version_info[0] == 2:
            # I _really_ hate doing this, but I can't figure out a
            # better way to handle unicode such that it works both
            # in python 2 and 3. There must be a better way, but
            # my unicode fu is weak.
            message = message.encode('utf-8')

        print(self.args.format.format(linenumber=linenumber, filename=filename,
                                      severity=severity, message=message,
                                      rulename=rulename, char=char))

    def _get_rules(self, cls):
        """Returns a list of rules of a given class

        Rules are treated as singletons - we only instantiate each
        rule once.
        """

        result = []
        for rule_class in cls.__subclasses__():
            rule_name = rule_class.__name__.lower()
            if rule_name not in self._rules:
                rule = rule_class(self)
                self._rules[rule_name] = rule
            result.append(self._rules[rule_name])
        return result

    def _load_rule_file(self, filename):
        """Import the given rule file"""
        abspath = os.path.abspath(filename)
        if abspath in IMPORTED_RULE_FILES:
            return
        if not (os.path.exists(filename)):
            sys.stderr.write("rflint: %s: No such file or directory\n" % filename)
            return
        try:
            basename = os.path.basename(filename)
            (name, ext) = os.path.splitext(basename)
            imp.load_source(name, filename)
            IMPORTED_RULE_FILES.append(abspath)
        except Exception as e:
            sys.stderr.write("rflint: %s: exception while loading: %s\n" % (filename, str(e)))

    def parse_and_process_args(self, args):
        """Handle the parsing of command line arguments."""

        parser = argparse.ArgumentParser(
            prog="python -m rflint",
            description="A static analyzer for robot framework plain text files.",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog = (
                "You can use 'all' in place of RULENAME to refer to all rules. \n"
                "\n"
                "For example: '--ignore all --warn DuplicateTestNames' will ignore all\n"
                "rules except DuplicateTestNames.\n"
                "\n"
                "FORMAT is a string that performs a substitution on the following \n"
                "patterns: {severity}, {linenumber}, {char}, {message}, and {rulename}.\n"
                "\n"
                "For example: --format 'line: {linenumber}: message: {message}'. \n"
                "\n"
                "ARGUMENTFILE is a filename with contents that match the format of \n"
                "standard robot framework argument files\n"
                "\n"
                "If you give a directory as an argument, all files in the directory\n"
                "with the suffix .txt, .robot, .resource, or .tsv will be processed. \n"
                "With the --recursive option, subfolders within the directory will \n"
                "also be processed."
                )
            )
        parser.add_argument("--error", "-e", metavar="RULENAME", action=SetErrorAction,
                            help="Assign a severity of ERROR to the given RULENAME")
        parser.add_argument("--ignore", "-i", metavar="RULENAME", action=SetIgnoreAction,
                            help="Ignore the given RULENAME")
        parser.add_argument("--warning", "-w", metavar="RULENAME", action=SetWarningAction,
                            help="Assign a severity of WARNING for the given RULENAME")
        parser.add_argument("--list", "-l", action="store_true",
                            help="show a list of known rules and exit")
        parser.add_argument("--describe", "-d", action="store_true",
                            help="describe the given rules")
        parser.add_argument("--no-filenames", action="store_false", dest="print_filenames",
                            default=True,
                            help="suppress the printing of filenames")
        parser.add_argument("--format", "-f",
                            help="Define the output format",
                            default='{severity}: {linenumber}, {char}: {message} ({rulename})')
        parser.add_argument("--version", action="store_true", default=False,
                            help="Display version number and exit")
        parser.add_argument("--verbose", "-v", action="store_true", default=False,
                            help="Give verbose output")
        parser.add_argument("--configure", "-c", action=ConfigureAction,
                            help="Configure a rule")
        parser.add_argument("--recursive", "-r", action="store_true", default=False,
                            help="Recursively scan subfolders in a directory")
        parser.add_argument("--rulefile", "-R", action=RulefileAction,
                            help="import additional rules from the given RULEFILE")
        parser.add_argument("--argumentfile", "-A", action=ArgfileLoader,
                            help="read arguments from the given file")
        parser.add_argument('args', metavar="file", nargs=argparse.REMAINDER)

        # create a custom namespace, in which we can store a reference to
        # our rules. This lets the custom argument actions access the list
        # of rules
        ns = argparse.Namespace()
        setattr(ns, "app", self)
        args = parser.parse_args(args, ns)

        Rule.output_format = args.format

        return args


class RulefileAction(argparse.Action):
    def __call__(self, parser, namespace, arg, option_string=None):
        app = getattr(namespace, "app")
        app._load_rule_file(arg)


class ConfigureAction(argparse.Action):
    def __call__(self, parser, namespace, arg, option_string=None):
        rulename, argstring = arg.split(":", 1)
        args = argstring.split(":")
        app = getattr(namespace, "app")

        for rule in app.all_rules:
            if rulename == rule.name:
                rule.configure(*args)
                return
        raise UnknownRuleException(rulename)


class SetStatusAction(argparse.Action):
    """Abstract class which provides a method for checking the rule name"""
    def check_rule_name(self, rulename, rules):
        if (rulename != "all" and
            rulename.lower() not in [rule.name.lower() for rule in rules]):
            raise UnknownRuleException(rulename)


class SetWarningAction(SetStatusAction):
    """Called when the argument parser encounters --warning"""
    def __call__(self, parser, namespace, rulename, option_string = None):

        app = getattr(namespace, "app")
        self.check_rule_name(rulename, app.all_rules)

        for rule in app.all_rules:
            if rulename == rule.name or rulename == "all":
                rule.severity = WARNING


class SetErrorAction(SetStatusAction):
    """Called when the argument parser encounters --error"""
    def __call__(self, parser, namespace, rulename, option_string = None):

        app = getattr(namespace, "app")
        self.check_rule_name(rulename, app.all_rules)

        for rule in app.all_rules:
            if rulename == rule.name or rulename == "all":
                rule.severity = ERROR


class SetIgnoreAction(SetStatusAction):
    """Called when the argument parser encounters --ignore"""
    def __call__(self, parser, namespace, rulename, option_string = None):

        app = getattr(namespace, "app")
        self.check_rule_name(rulename, app.all_rules)

        for rule in app.all_rules:
            if rulename == rule.name or rulename == "all":
                rule.severity = IGNORE


class ArgfileLoader(argparse.Action):
    """Called when the argument parser encounters --argumentfile"""
    def __call__ (self, parser, namespace, values, option_string = None):
        ap = ArgFileParser(["--argumentfile","-A"])
        args = ap.process(["-A", values])
        parser.parse_args(args, namespace)



'''
Copyright 2014 Bryan Oakley

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from rflint.common import KeywordRule, ERROR
from rflint.parser import SettingTable

class RequireKeywordDocumentation(KeywordRule):
    '''Verify that a keyword has documentation'''
    severity=ERROR

    def apply(self, keyword):
        for setting in keyword.settings:
            if setting[1].lower() == "[documentation]" and len(setting) > 2:
                return

        # set the line number to the line immediately after the keyword name
        self.report(keyword, "No keyword documentation", keyword.linenumber+1)


class TooFewKeywordSteps(KeywordRule):
    '''Keywords should have at least a minimum number of steps

    This rule is configurable. The default number of required steps is 2.
    '''

    min_required = 2

    def configure(self, min_required):
        self.min_required = int(min_required)

    def apply(self, keyword):
        # ignore empty steps
        steps = [step for step in keyword.steps if not (len(step) == 1 and not step[0])]
        if len(steps) < self.min_required:
            msg = "Too few steps (%s) in keyword" % len(steps)
            self.report(keyword, msg, keyword.linenumber)



from __future__ import print_function

from .common import Statement

class RobotTable(object):
    '''A table made up of zero or more rows'''
    def __init__(self, parent, linenumber=0, name=None, header=None):
        self.linenumber = linenumber
        self.name = name
        self.rows = []
        self.comments = []
        self.parent = parent
        self.header = header

    def dump(self):
        for row in self.rows:
            print("| " + " | ".join(row))

    def append(self, row):
        self.rows.append(row)

    def __str__(self):
        if self.name is None:
            return "None"
        else:
            return self.name

    def __repr__(self):
        return "<%s(linenumbmer=%s, name=\"%s\")>" % (self.__class__.__name__, self.linenumber, self.name)

class SimpleTableMixin(object):
    '''Mixin to handle simple tables (tables other than keywords and tests)'''

    @property
    def statements(self):
        '''Return a list of statements

        This is done by joining together any rows that
        have continuations
        '''
        # FIXME: no need to do this every time; we should cache the
        # result
        if len(self.rows) == 0:
            return []

        current_statement = Statement(self.rows[0])
        current_statement.startline = self.rows[0].linenumber
        current_statement.endline = self.rows[0].linenumber
        statements = []
        for row in self.rows[1:]:
            if len(row) > 0 and row[0] == "...":
                # we found a continuation
                current_statement += row[1:]
                current_statement.endline = row.linenumber
            else:
                if len(current_statement) > 0:
                    # append current statement to the list of statements...
                    statements.append(current_statement)
                # start a new statement
                current_statement = Statement(row)
                current_statement.startline = row.linenumber
                current_statement.endline = row.linenumber

        if len(current_statement) > 0:
            statements.append(current_statement)

        # trim trailing blank statements
        while (len(statements[-1]) == 0 or 
               ((len(statements[-1]) == 1) and len(statements[-1][0]) == 0)):
            statements.pop()
        return statements

class DefaultTable(RobotTable): pass  # the table with no name
class UnknownTable(RobotTable): pass  # a table with an unknown header
class SettingTable(RobotTable, SimpleTableMixin): pass
class VariableTable(RobotTable): pass
class MetadataTable(RobotTable): pass

class AbstractContainerTable(RobotTable):
    '''Parent class of Keyword and Testcase tables'''
    _childClass = None
    def __init__(self, parent, *args, **kwargs):
        if self._childClass is None:
            # hey! Don't try to instantiate this class directly.
            raise Exception("D'oh! This is an abstract class.")
        super(AbstractContainerTable, self).__init__(parent, *args, **kwargs)
        self._children = []
        self.parent = parent

    def dump(self):
        for child in self._children:
            print("| " + child.name)
            for row in child.rows:
                row.dump()

    def append(self, row):
        ''' 
        The idea is, we recognize when we have a new testcase by 
        checking the first cell. If it's not empty and not a comment, 
        we have a new test case.

        '''
        if len(row) == 0:
            # blank line. Should we throw it away, or append a BlankLine object?
            return

        if (row[0] != "" and 
            (not row[0].lstrip().startswith("#"))):
            # we have a new child table
            self._children.append(self._childClass(self.parent, row.linenumber, row[0]))
            if len(row.cells) > 1:
                # It appears the first row -- which contains the test case or
                # keyword name -- also has the first logical row of cells.
                # We'll create a Row, but we'll make the first cell empty instead
                # of leaving the name in it, since other code always assumes the
                # first cell is empty. 
                #
                # To be honest, I'm not sure this is the Right Thing To Do, but 
                # I'm too lazy to audit the code to see if it matters if we keep 
                # the first cell intact. Sorry if this ends up causing you grief
                # some day...
                row[0] = ""
                self._children[-1].append(row.linenumber, row.raw_text, row.cells)

        elif len(self._children) == 0:
            # something before the first test case
            # For now, append it to self.comments; eventually we should flag
            # an error if it's NOT a comment
            self.comments.append(row)

        else:
            # another row for the testcase
            if len(row.cells) > 0:
                self._children[-1].append(row.linenumber, row.raw_text, row.cells)





from rflint.common import SuiteRule, ERROR, WARNING, normalize_name
from rflint.parser import SettingTable
import re

class PeriodInSuiteName(SuiteRule):
    '''Warn about periods in the suite name

    Since robot uses "." as a path separator, using a "." in a suite
    name can lead to ambiguity.
    '''
    severity = WARNING

    def apply(self,suite):
        if "." in suite.name:
            self.report(suite, "'.' in suite name '%s'" % suite.name, 0)

class InvalidTable(SuiteRule):
    '''Verify that there are no invalid table headers'''
    severity = WARNING

    def apply(self, suite):
        for table in suite.tables:
            if (not re.match(r'^(comments?|settings?|metadata|(test )?cases?|(user )?keywords?|variables?)$', 
                             table.name, re.IGNORECASE)):
                self.report(suite, "Unknown table name '%s'" % table.name, table.linenumber)


class DuplicateKeywordNames(SuiteRule):
    '''Verify that no keywords have a name of an existing keyword in the same file'''
    severity = ERROR

    def apply(self, suite):
        cache = []
        for keyword in suite.keywords:
            # normalize the name, so we catch things like
            # Smoke Test vs Smoke_Test, vs SmokeTest, which
            # robot thinks are all the same
            name = normalize_name(keyword.name)
            if name in cache:
                self.report(suite, "Duplicate keyword name '%s'" % keyword.name, keyword.linenumber)
            cache.append(name)

class DuplicateTestNames(SuiteRule):
    '''Verify that no tests have a name of an existing test in the same suite'''
    severity = ERROR

    def apply(self, suite):
        cache = []
        for testcase in suite.testcases:
            # normalize the name, so we catch things like
            # Smoke Test vs Smoke_Test, vs SmokeTest, which
            # robot thinks are all the same
            name = normalize_name(testcase.name)
            if name in cache:
                self.report(suite, "Duplicate testcase name '%s'" % testcase.name, testcase.linenumber)
            cache.append(name)

class RequireSuiteDocumentation(SuiteRule):
    '''Verify that a test suite has documentation'''
    severity=WARNING

    def apply(self, suite):
        for table in suite.tables:
            if isinstance(table, SettingTable):
                for row in table.rows:
                    if row[0].lower() == "documentation":
                        return
        # we never found documentation; find the first line of the first
        # settings table, default to the first line of the file
        linenum = 1
        for table in suite.tables:
            if isinstance(table, SettingTable):
                linenum = table.linenumber + 1
                break

        self.report(suite, "No suite documentation", linenum)

class TooManyTestCases(SuiteRule):
    '''
    Should not have too many tests in one suite.

    The exception is if they are data-driven.

    https://code.google.com/p/robotframework/wiki/HowToWriteGoodTestCases#Test_suite_structure

    You can configure the maximum number of tests. The default is 10.
    '''
    severity = WARNING
    max_allowed = 10

    def configure(self, max_allowed):
        self.max_allowed = int(max_allowed)

    def apply(self, suite):
        # check for template (data-driven tests)
        for table in suite.tables:
            if isinstance(table, SettingTable):
                for row in table.rows:
                    if row[0].lower() == "test template":
                        return
        # we didn't find a template, so these aren't data-driven
        testcases = list(suite.testcases)
        if len(testcases) > self.max_allowed:
            self.report(
                suite, "Too many test cases (%s > %s) in test suite"
                % (len(testcases), self.max_allowed), testcases[self.max_allowed].linenumber
            )



from rflint.common import SuiteRule, ResourceRule, ERROR, normalize_name

def check_duplicates(report_duplicate, table,
                     permitted_dups=None, normalize_itemname=normalize_name):
    # `table` is a SettingsTable or a VariableTable; either contains rows,
    # but only VariableTable also contains statements.
    seen_rows = {}
    for row in table.rows:
        item = normalize_itemname(row[0])

        # skip empty lines, comments and continuation lines
        if item == "":
            continue
        if item.startswith("#"):
            continue
        if item.startswith("..."):
            continue

        # some tables allow duplicates
        if permitted_dups and item in permitted_dups:
            continue

        if item in seen_rows:
            prev_row = seen_rows[item]
            report_duplicate(row, prev_row)
        else:
            seen_rows[item] = row


class DuplicateSettingsCommon(object):
    '''Verify that settings are not repeated in a Settings table

    This has been made an error in Robot3.0
    https://github.com/robotframework/robotframework/issues/2204'''
    severity = ERROR

    def apply(self, suite):
        def report_duplicate_setting(setting, prev_setting):
            self.report(suite,
                "Setting '%s' used multiple times (previously used line %d)" % \
                (setting[0], prev_setting.linenumber), setting.linenumber)

        for table in suite.tables:
            if table.name == "Settings":
                check_duplicates(report_duplicate_setting, table,
                    permitted_dups=["library", "resource", "variables"])

class DuplicateSettingsInSuite(DuplicateSettingsCommon, SuiteRule):
    pass

class DuplicateSettingsInResource(DuplicateSettingsCommon, ResourceRule):
    pass


def strip_variable_name(varname):
    return varname.lstrip("${").rstrip("}= ")

def normalize_variable_name(varname):
    return normalize_name(strip_variable_name(varname))

class DuplicateVariablesCommon(object):
    '''Verify that variables are not defined twice in the same table

    This is not an error, but leads to surprising result (first definition
    wins, later is ignored).'''
    def apply(self, suite):
        def report_duplicate_variable(variable, prev_variable):
            self.report(suite,
                "Variable '%s' defined twice, previous definition line %d" % \
                (strip_variable_name(variable[0]), prev_variable.linenumber),
                variable.linenumber)

        for table in suite.tables:
            if table.name == "Variables":
                check_duplicates(report_duplicate_variable, table,
                                 normalize_itemname=normalize_variable_name)

class DuplicateVariablesInSuite(DuplicateVariablesCommon, SuiteRule):
    pass

class DuplicateVariablesInResource(DuplicateVariablesCommon, ResourceRule):
    pass



'''
Copyright 2014 Bryan Oakley

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
'''

from rflint.common import TestRule, ERROR, WARNING
from rflint.parser import SettingTable


class PeriodInTestName(TestRule):
    '''Warn about periods in the testcase name
    
    Since robot uses "." as a path separator, using a "." in a testcase
    name can lead to ambiguity. 
    '''
    severity = WARNING
    
    def apply(self,testcase):
        if "." in testcase.name:
            self.report(testcase, "'.' in testcase name '%s'" % testcase.name, testcase.linenumber)

class TagWithSpaces(TestRule):
    '''Flags tags that have spaces in the tag name'''
    severity=ERROR

    def apply(self, testcase):
        for tag in testcase.tags:
            if ((" " in tag) or ("\t" in tag)):
                self.report(testcase, "space not allowed in tag name: '%s'" % tag, testcase.linenumber)

class RequireTestDocumentation(TestRule):
    '''Verify that a test suite has documentation

    This rule is not enforced for data driven tests ("Test Template" in Settings)
    '''
    severity=ERROR

    def apply(self, testcase):
        if testcase.is_templated:
            return

        for setting in testcase.settings:
            if setting[1].lower() == "[documentation]" and len(setting) > 2:
                return

        # set the line number to the line immediately after the testcase name
        self.report(testcase, "No testcase documentation", testcase.linenumber+1)

class TooFewTestSteps(TestRule):
    '''Tests should have at least a minimum number of steps

    This rule is configurable. The default number of required steps is 2.
    '''

    min_required = 2

    def configure(self, min_required):
        self.min_required = int(min_required)

    def apply(self, testcase):
        if testcase.is_templated:
            return

        # ignore empty test steps
        steps = [step for step in testcase.steps if not (len(step) == 1
                                                         and not step[0])]
        if len(steps) < self.min_required:
            msg = "Too few steps (%s) in test case" % len(steps)
            self.report(testcase, msg,  testcase.linenumber)

class TooManyTestSteps(TestRule):
    '''Workflow tests should have no more than ten steps.

    https://code.google.com/p/robotframework/wiki/HowToWriteGoodTestCases#Workflow_tests
    '''

    severity=WARNING
    max_allowed = 10

    def configure(self, max_allowed):
        self.max_allowed = int(max_allowed)

    def apply(self, testcase):
        if testcase.is_templated:
            return

        # ignore empty test steps
        steps = [step for step in testcase.steps if not (len(step) == 1
                                                         and not step[0])]
        if len(steps) > self.max_allowed:
            self.report(testcase,
                        "Too many steps (%s) in test case" % len(steps),
                        steps[self.max_allowed].startline)



from __future__ import print_function

import time
import re

class Matcher(object):
    '''A convenience class for regular expression matching

    Example:
        match = Matcher(re.IGNORECASE)
        if match(pattern1, string):
            print match.group(1)
        elif match(pattern2, string):
            print match.group(1)
        ...
    '''

    def __init__(self, flags=None):
        self.flags = flags

    def __call__(self, pattern, string, flags=None):
        if flags is None:
            self.result = re.match(pattern, string, flags=self.flags)
        else:
            self.result = re.match(pattern, string)
        return self.result

    def __getattr__(self, attr):
        return getattr(self.result, attr)


def timeit(func):
    def wrapper(*arg, **kw):
        '''source: http://www.daniweb.com/code/snippet368.html'''
        t1 = time.time()
        res = func(*arg, **kw)
        t2 = time.time()
        print(func.__name__, ":", int(1000*(t2-t1)), "ms")
        return res
    return wrapper




from __future__ import print_function
import re

class RobotStatements(object):
    def append(self, linenumber, raw_text, cells):
        """Add another row of data from a test suite"""
        self.rows.append(Row(linenumber, raw_text, cells))

    @property
    def path(self):
        # this property exists so that the linter doesn't
        # have to have this logic
        return self.parent.path

    @property
    def steps(self):
        """Return a list of steps (statements that are not settings or comments)"""
        steps = []
        for statement in self.statements:
            if ((not statement.is_comment()) and
                (not statement.is_setting())):
                steps.append(statement)
        return steps

    @property
    def settings(self):
        """Return a list of settings (statements with cell[1] matching \[.*?\])

        Note: this returns any statement that *looks* like a setting. If you have
        a misspelled or completely bogus setting, it'll return that too
        (eg: | | [Blockumentation] | hello, world)
        """
        return [statement for statement in self.statements
                if (statement.is_setting() and not statement.is_comment())]

    @property
    def statements(self):
        """Return a list of statements

        This is done by joining together any rows that
        have continuations
        """
        # FIXME: no need to do this every time; we should cache the
        # result
        if len(self.rows) == 0:
            return []

        current_statement = Statement(self.rows[0])
        current_statement.startline = self.rows[0].linenumber
        current_statement.endline = self.rows[0].linenumber
        statements = []
        for row in self.rows[1:]:
            if len(row) > 1 and row[0] == "" and row[1] == "...":
                # we found a continuation
                current_statement += row[2:]
                current_statement.endline = row.linenumber
            else:
                if len(current_statement) > 0:
                    # append current statement to the list of statements...
                    statements.append(current_statement)
                # start a new statement
                current_statement = Statement(row)
                current_statement.startline = row.linenumber
                current_statement.endline = row.linenumber

        if len(current_statement) > 0:
            statements.append(current_statement)

        return statements


# TODO: make Row and Statement more similar -- either
# both should inherit from list, or neither should.
class Row(object):
    """A row is made up of a list of cells plus metadata"""
    def __init__(self, linenumber, raw_text, cells):
        self.linenumber = linenumber
        self.raw_text = raw_text
        self.cells = cells

    def dump(self):
        print("|" + " | ".join([cell.strip() for cell in self.cells]))
    def __len__(self):
        return len(self.cells)
    def __setitem__(self, key, value):
        self.cells[key] = value
        return self.cells[key]
    def __getitem__(self, key):
        return self.cells[key]
    def __repr__(self):
        return "<line: %s cells: %s>" % (self.linenumber, str(self.cells))
    def __contains__(self, key):
        return key in self.cells

class Comment(Row):
    # this isn't entirely correct or well thought out.
    # I need a way to capture comments rather than
    # throw them away (mainly so I can recreate the original
    # file from the parsed data)
    pass

class Statement(list):
    """A Statement is a list of cells, plus some metadata"""
    startline = None
    endline = None

    def is_setting(self):
        if ((len(self) > 1) and
            (re.match(r'\[.*?\]', self[1]))):
            return True
        return False

    def is_comment(self):
        '''Return True if the first non-empty cell starts with "#"'''

        for cell in self[:]:
            if cell == "":
                continue

            # this is the first non-empty cell. Check whether it is
            # a comment or not.
            if cell.lstrip().startswith("#"):
                return True
            else:
                return False
        return False

    def __repr__(self):
        return "(%.4s-%.4s)%s" % (self.startline, self.endline, list.__repr__(self))



from .parser import (SuiteFolder, ResourceFile, SuiteFile, RobotFactory, 
                     Testcase, Keyword, Row, Statement, TestcaseTable, KeywordTable)
from .tables import DefaultTable, SettingTable, UnknownTable, VariableTable, MetadataTable, RobotTable




from .rflint import RfLint, ERROR, WARNING
from .parser import RobotFactory, SuiteFile, ResourceFile, Testcase, Keyword
from .common import SuiteRule, ResourceRule, TestRule, KeywordRule, GeneralRule
from .version import __version__



from .tables import SettingTable
from .common import Row, Statement, RobotStatements
import re


class Testcase(RobotStatements):
    def __init__(self, parent, linenumber, name):
        RobotStatements.__init__(self)
        self.linenumber = linenumber
        self.name = name
        self.rows = []
        self.parent = parent

    @property
    def is_templated(self):
        """Return True if the test is part of a suite that uses a Test Template"""
        for table in self.parent.tables:
            if isinstance(table, SettingTable):
                for row in table.rows:
                    if row[0].lower() == "test template":
                        return True
        return False

    # this is great, except that we don't return the line number
    # or character position of each tag. The linter needs that. :-(
    @property
    def tags(self):
        tags = []
        for statement in self.statements:
            if len(statement) > 2 and statement[1].lower() == "[tags]":
                tags = tags + statement[2:]
        return tags

    def __repr__(self):
        # should this return the fully qualified name?
        return "<Testcase: %s>" % self.name



from rflint.common import TestRule, KeywordRule, GeneralRule, ERROR, WARNING

import re


class LineTooLong(GeneralRule):
    '''Check that a line is not too long (configurable; default=100)'''

    severity = WARNING
    maxchars = 100

    def configure(self, maxchars):
        self.maxchars = int(maxchars)

    def apply(self, robot_file):
        for linenumber, line in enumerate(robot_file.raw_text.split("\n")):
            if len(line) > self.maxchars:
                message = "Line is too long (exceeds %s characters)" % self.maxchars
                self.report(robot_file, message, linenumber+1, self.maxchars)

class TrailingBlankLines(GeneralRule):
    '''Check for multiple blank lines at the end of a file

    This is a configurable. The default value is 2.
    '''

    severity = WARNING
    max_allowed = 2

    def configure(self, max_allowed):
        self.max_allowed=int(max_allowed)

    def apply(self, robot_file):
        # I realize I'm making two full passes over the data, but
        # python is plenty fast enough. Even processing a file with
        # over six thousand lines, this takes a couple of
        # milliseconds.  Plenty fast enough for the intended use case,
        # since most files should be about two orders of magnitude
        # smaller than that.

        match=re.search(r'(\s*)$', robot_file.raw_text)
        if match:
            count = len(re.findall(r'\n', match.group(0)))
            if count > self.max_allowed:
                numlines = len(robot_file.raw_text.split("\n"))
                message = "Too many trailing blank lines"
                linenumber = numlines-count
                self.report(robot_file, message, linenumber+self.max_allowed, 0)


class TrailingWhitespace(GeneralRule):
    severity = WARNING

    def apply(self, robot_file):
        for linenumber, line in enumerate(robot_file.raw_text.splitlines()):
            if len(line) != len(line.rstrip()):
                message = "Line has trailing whitespace"
                self.report(robot_file, message, linenumber+1)


class FileTooLong(GeneralRule):
    '''Verify the file has fewer lines than a given threshold.

    You can configure the maximum number of lines. The default is 300.
    '''

    severity = WARNING
    max_allowed = 300

    def configure(self, max_allowed):
        self.max_allowed = int(max_allowed)

    def apply(self, robot_file):
        lines = robot_file.raw_text.split("\n")
        if len(lines) > self.max_allowed:
            message = "File has too many lines (%s)" % len(lines)
            linenumber = self.max_allowed+1
            self.report(robot_file, message, linenumber, 0)



import sys
from .rflint import RfLint


def main(args=None):
    try:
        app = RfLint()
        result = app.run(args)
        return result

    except Exception as e:
        sys.stderr.write(str(e) + "\n")
        return 1

if __name__ == "__main__":
    if len(sys.argv) == 1:
        sys.argv.append("--help")
    sys.exit(main(sys.argv[1:]))



'''
This file defines custom exceptions used by robotframework-lint

'''

class UnknownRuleException(Exception):
    def __init__(self, rulename):
        super(UnknownRuleException, self).__init__("unknown rule: '%s'" % rulename)



# Does this need to be in its own file, or can I combine 
# testcase and keyword into one file? 
from .common import RobotStatements

class Keyword(RobotStatements):
    '''A robotframework keyword

    A keyword is identical to a testcase in almost all respects
    except for some of the metadata it supports (which this definition
    doesn't (yet) account for...).
    '''
    def __init__(self, parent, linenumber, name):
        RobotStatements.__init__(self)
        self.linenumber = linenumber
        self.name = name
        self.rows = []
        self.parent = parent

    def __repr__(self):
        # should this return the fully qualified name?
        return "<Keyword: %s>" % self.name




import sys

ERROR = "E"
WARNING = "W"
IGNORE = "I"


def normalize_name(string):
    '''convert to lowercase, remove spaces and underscores'''
    return string.replace(" ", "").replace("_", "").lower()


class Rule(object): 
    # default severity; subclasses may override
    severity = WARNING
    output_format = "{severity}: {linenumber}, {char}: {message} ({rulename})"

    def __init__(self, controller, severity=None):
        self.controller = controller
        if severity is not None:
            self.severity = severity

    def configure(self, *args, **kwargs):
        # subclasses can override this if they want to be configurable.
        raise Exception("rule '%s' is not configurable" % self.__class__.__name__)

    @property
    def name(self):
        return self.__class__.__name__

    def report(self, obj, message, linenum, char_offset=0):
        """Report an error or warning"""
        self.controller.report(linenumber=linenum, filename=obj.path,
                               severity=self.severity, message=message,
                               rulename = self.__class__.__name__,
                               char=char_offset)

    @property
    def doc(self):
        '''Algorithm from https://www.python.org/dev/peps/pep-0257/'''
        if not self.__doc__:
            return ""

        lines = self.__doc__.expandtabs().splitlines()

        # Determine minimum indentation (first line doesn't count):
        indent = sys.maxsize
        for line in lines[1:]:
            stripped = line.lstrip()
            if stripped:
                indent = min(indent, len(line) - len(stripped))

        # Remove indentation (first line is special):
        trimmed = [lines[0].strip()]
        if indent < sys.maxsize:
            for line in lines[1:]:
                trimmed.append(line[indent:].rstrip())

        # Strip off trailing and leading blank lines:
        while trimmed and not trimmed[-1]:
            trimmed.pop()
        while trimmed and not trimmed[0]:
            trimmed.pop(0)

        # Return a single string:
        return '\n'.join(trimmed)

    def __repr__(self):
        return "%s %s" % (self.severity, self.__class__.__name__)

class TestRule(Rule): 
    """Rule that runs against test cases. 

    The object that is passed in will be of type rflint.parser.Testcase
    """
    pass

class ResourceRule(Rule):
    """Rule that runs against a resource file

    The object that is passed in will be of type rflint.parser.ResourceFile
    """

class SuiteRule(Rule): 
    """Rule that runs against test cases. 

    The object that is passed in will be of type rflint.parser.SuiteFile
    """
    pass

class KeywordRule(Rule): 
    """Rule that runs against keywords

    The object that is passed in will be of type rflint.parser.Keyword
    """
    pass

class GeneralRule(Rule):
    """Rule that requires a suite or resource file, but may apply to child objects

    This rule is identical to a SuiteRule, but exists in case you want
    to write a rule that accepts a suite but doesn't necessarily apply
    to the suite (ie: you may iterate over tests, or keywords, or some
    other child object)
    """
    pass




# N.B. to push a new version to PyPi, update the version number
# in rflint/version.py and then run 'python setup.py sdist upload'

from setuptools import setup

filename = 'rflint/version.py'
exec(compile(open(filename, 'rb').read(), filename, 'exec'))

setup(
    name             = 'rf-lint',
    version          = __version__,
    author           = 'Bryan Oakley',
    author_email     = 'bryan.oakley@gmail.com',
    url              = 'https://github.com/boakley/robotframework-lint/',
    keywords         = 'robotframework',
    license          = 'Apache License 2.0',
    description      = 'Static analysis tool for robotframework plain text files',
    long_description = open('README.md').read(),
    zip_safe         = False,
    include_package_data = True,
    install_requires = ['robotframework'],
    classifiers      = [
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Framework :: Robot Framework",
        "Programming Language :: Python",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Quality Assurance",
        "Intended Audience :: Developers",
        "Environment :: Console",
        ],
    packages         =[
        'rflint',
        'rflint.rules',
        'rflint.parser',
        ],
    scripts          =[],
    entry_points={
        'console_scripts': [
            "rflint = rflint.__main__:main"
        ]
    }
)



'''

A custom robotframework parser that retains line numbers (though
it doesn't (yet!) retain character positions for each cell)


Note: this only works on pipe and space separated files. It uses part
of the TxtReader robot parser to divide a line into cells.

(probably works for space-separated too. I haven't tried. )

Performance is pretty spiffy! At the time I write this (where
admittedly I don't fully parse everything) it is about 3x-5x faster
than the official robot parser. It can read a file with 500
test cases and 500 keywords in about 30ms, compared to 150ms
for the robot parser. Sweet.

'''
from __future__ import print_function

import re
import sys
import os.path
from robot.parsing.txtreader import TxtReader
from robot.errors import DataError
from robot.utils import Utf8Reader
from .util import timeit, Matcher
from .tables import AbstractContainerTable, DefaultTable, SettingTable, VariableTable, UnknownTable
from .testcase import Testcase
from .rfkeyword import Keyword
from .common import Row, Statement


def RobotFactory(path, parent=None):
    '''Return an instance of SuiteFile, ResourceFile, SuiteFolder

    Exactly which is returned depends on whether it's a file or
    folder, and if a file, the contents of the file. If there is a
    testcase table, this will return an instance of SuiteFile,
    otherwise it will return an instance of ResourceFile.
    '''

    if os.path.isdir(path):
        return SuiteFolder(path, parent)

    else:
        rf = RobotFile(path, parent)

        for table in rf.tables:
            if isinstance(table, TestcaseTable):
                rf.__class__ = SuiteFile
                return rf

        rf.__class__ = ResourceFile
        return rf

class SuiteFolder(object):
    def __init__(self, path, parent=None):

        self.path = os.path.abspath(path)
        self.parent = parent
        self.name = os.path.splitext(os.path.basename(path))[0]
        self.initfile = None

        # see if there's an initialization file. If so,
        # attempt to load it
        for filename in ("__init__.robot", "__init__.txt"):
            if os.path.exists(os.path.join(self.path, filename)):
                self.initfile = RobotFile(os.path.join(self.path, filename))
                break


    def walk(self, *types):
        '''
        Iterator which visits all suites and suite files,
        yielding test cases and keywords
        '''
        requested = types if len(types) > 0 else [SuiteFile, ResourceFile, SuiteFolder, Testcase, Keyword]

        for thing in self.robot_files:
            if thing.__class__ in requested:
                yield thing
            if isinstance(thing, SuiteFolder):
                for child in thing.walk():
                    if child.__class__ in requested:
                        yield child
            else:
                for child in thing.walk(*types):
                    yield child

    @property
    def robot_files(self):
        '''Return a list of all folders, and test suite files (.txt, .robot)
        '''
        result = []
        for name in os.listdir(self.path):
            fullpath = os.path.join(self.path, name)
            if os.path.isdir(fullpath):
                result.append(RobotFactory(fullpath, parent=self))
            else:
                if ((name.endswith(".txt") or name.endswith(".robot")) and
                    (name not in ("__init__.txt", "__init__.robot"))):

                    result.append(RobotFactory(fullpath, parent=self))
        return result


class RobotFile(object):
    '''
    Terminology:

    - A file is a set of tables
    - A table begins with a heading and extends to the next table or EOF
    - Each table may be made up of smaller tables that define test cases
      or keywords
    - Each line of text in a table becomes a "Row".
    - A Row object contains a list of cells.
    - A cell is all of the data between pipes, stripped of leading and
      trailing spaces

    '''
    def __init__(self, path, parent=None):
        self.parent = parent
        self.name = os.path.splitext(os.path.basename(path))[0]
        self.path = os.path.abspath(path)
        self.tables = []
        self.rows = []

        try:
            self._load(path)
        except Exception as e:
            sys.stderr.write("there was a problem reading '%s': %s\n" % (path, str(e)))

    def walk(self, *types):
        '''
        Iterator which can return all test cases and/or keywords

        You can specify with objects to return as parameters; if
        no parameters are given, both tests and keywords will
        be returned.

        For example, to get only test cases, you could call it
        like this:

            robot_file = RobotFactory(...)
            for testcase in robot_file.walk(Testcase): ...

        '''
        requested = types if len(types) > 0 else [Testcase, Keyword]

        if Testcase in requested:
            for testcase in self.testcases:
                yield testcase

        if Keyword in requested:
            for keyword in self.keywords:
                yield keyword

    def _load(self, path):

        '''
        The general idea is to do a quick parse, creating a list of
        tables. Each table is nothing more than a list of rows, with
        each row being a list of cells. Additional parsing such as
        combining rows into statements is done on demand. This first
        pass is solely to read in the plain text and organize it by table.
        '''

        self.tables = []
        current_table = DefaultTable(self)

        with Utf8Reader(path) as f:
            # N.B. the caller should be catching errors
            self.raw_text = f.read()

            f._file.seek(0) # bleh; wish this wasn't a private property
            matcher = Matcher(re.IGNORECASE)
            for linenumber, raw_text in enumerate(f.readlines()):
                linenumber += 1; # start counting at 1 rather than zero

                # this mimics what the robot TSV reader does --
                # it replaces non-breaking spaces with regular spaces,
                # and then strips trailing whitespace
                raw_text = raw_text.replace(u'\xA0', ' ')
                raw_text = raw_text.rstrip()

                # FIXME: I'm keeping line numbers but throwing away
                # where each cell starts. I should be preserving that
                # (though to be fair, robot is throwing that away so
                # I'll have to write my own splitter if I want to save
                # the character position)
                cells = TxtReader.split_row(raw_text)
                _heading_regex = r'^\s*\*+\s*(.*?)[ *]*$'

                if matcher(_heading_regex, cells[0]):
                    # we've found the start of a new table
                    table_name = matcher.group(1)
                    current_table = tableFactory(self, linenumber, table_name, raw_text)
                    self.tables.append(current_table)
                else:
                    current_table.append(Row(linenumber, raw_text, cells))


    def __repr__(self):
        return "<RobotFile(%s)>" % self.path

    @property
    def type(self):
        '''Return 'suite' or 'resource' or None

        This will return 'suite' if a testcase table is found;
        It will return 'resource' if at least one robot table
        is found. If no tables are found it will return None
        '''

        robot_tables = [table for table in self.tables if not isinstance(table, UnknownTable)]
        if len(robot_tables) == 0:
            return None

        for table in self.tables:
            if isinstance(table, TestcaseTable):
                return "suite"

        return "resource"

    @property
    def keywords(self):
        '''Generator which returns all keywords in the suite'''
        for table in self.tables:
            if isinstance(table, KeywordTable):
                for keyword in table.keywords:
                    yield keyword

    @property
    def testcases(self):
        '''Generator which returns all test cases in the suite'''
        for table in self.tables:
            if isinstance(table, TestcaseTable):
                for testcase in table.testcases:
                    yield testcase

    def dump(self):
        '''Regurgitate the tables and rows'''
        for table in self.tables:
            print("*** %s ***" % table.name)
            table.dump()


def tableFactory(parent, linenumber, name, header):
    match = Matcher(re.IGNORECASE)
    if name is None:
        table = UnknownTable(parent, linenumber, name, header)
    elif match(r'settings?|metadata', name):
        table = SettingTable(parent, linenumber, name, header)
    elif match(r'variables?', name):
        table = VariableTable(parent, linenumber, name, header)
    elif match(r'test ?cases?', name):
        table = TestcaseTable(parent, linenumber, name, header)
    elif match(r'(user )?keywords?', name):
        table = KeywordTable(parent, linenumber, name, header)
    else:
        table = UnknownTable(parent, linenumber, name, header)

    return table


class SuiteFile(RobotFile):
    def __repr__(self):
        return "<SuiteFile(%s)>" % self.path

    @property
    def settings(self):
        '''Generator which returns all of the statements in all of the settings tables'''
        for table in self.tables:
            if isinstance(table, SettingTable):
                for statement in table.statements:
                    yield statement

    @property
    def variables(self):
        '''Generator which returns all of the statements in all of the variables tables'''
        for table in self.tables:
            if isinstance(table, VariableTable):
                # FIXME: settings have statements, variables have rows WTF? :-(
                for statement in table.rows:
                    if statement[0] != "":
                        yield statement

class ResourceFile(RobotFile):
    def __repr__(self):
        return "<ResourceFile(%s)>" % self.path

    @property
    def settings(self):
        '''Generator which returns all of the statements in all of the settings tables'''
        for table in self.tables:
            if isinstance(table, SettingTable):
                for statement in table.statements:
                    yield statement

class TestcaseTable(AbstractContainerTable):
    _childClass = Testcase
    def __init__(self, parent, *args, **kwargs):
        super(TestcaseTable, self).__init__(parent, *args, **kwargs)
        self.testcases = self._children

class KeywordTable(AbstractContainerTable):
    _childClass = Keyword
    def __init__(self, parent, *args, **kwargs):
        super(KeywordTable, self).__init__(parent, *args, **kwargs)
        self.keywords = self._children

@timeit
def dump(suite):
    result = []
    for table in suite.tables:
#        print "table:", table
#        for row in table.rows:
#            print "=>", row
        if isinstance(table, TestcaseTable):
            for tc in table.testcases:
                # force parsing of individual steps
                steps = [step for step in tc.steps]

if __name__ == "__main__":
    from robot.parsing import TestData, ResourceFile
    import sys

    # parse with the robot parser and this parser, to
    # see which is faster. Of course, this parser will
    # be faster :-)
    @timeit
    def test_robot():
        try:
            suite = TestData(parent=None, source=sys.argv[1])
        except DataError:
            # if loading the suite failed, assume it's a resource file
            # (bad assumption, but _whatever_)
            suite = ResourceFile(source=sys.argv[1])
        return suite

    @timeit
    def test_mine():
        suite = RobotFile(sys.argv[1])
        # force parsing of every line
        for tc in suite.testcases:
            statements = tc.statements
            tags = tc.tags
        return suite

    if len(sys.argv) == 1:
        print("give me a filename on the command line")
        sys.exit(1)

    suite1 = test_robot()
    suite2 = test_mine()
